/*
Module : NTRAY.CPP
Purpose: implementation for a MFC class to encapsulate Shell_NotifyIcon
Created: PJN / 14-05-1997
History: PJN / 25-11-1997 : Addition of the following
                            1. HideIcon(), ShowIcon() & MoveToExtremeRight 
                            2. Support for animated tray icons
         PJN / 23-06-1998 : Class now supports the new Taskbar Creation Notification 
                            message which comes with IE 4. This allows the tray icon
                            to be recreated whenever the explorer restarts (Crashes!!)
         PJN / 22-07-1998 : 1. Code now compiles cleanly at warning level 4
                            2. Code is now UNICODE enabled + build configurations are 
                            provided
                            3. The documentation for the class has been updated

         PJN / 27-01-1999 : Code first tries to load a 16*16 icon before loading the 32*32
                            version. This removes the blurryness which was previously occuring
         PJN / 28-01-1999 : Fixed a number of level 4 warnings which were occurring.
         PJN / 09-05-1999 : Fixed a problem as documented in KB article "PRB: Menus for 
                            Notification Icons Do Not Work Correctly", Article ID: Q135788 
         

Copyright (c) 1997 - 1999 by PJ Naughter.  
All rights reserved.

		SAW / 11-11-98		: ported to ATL by Shaun Wilde shaun.wilde@many-monkeys.freeserve.co.uk

		SAW / 06-05-99		: added selectable default menu item support.

		SAW / 26-05-99		: ported to 1.4 version of class
*/

/////////////////////////////////  Includes  //////////////////////////////////
#include "stdafx.h"
#include "resource.h"
#include "ntray.h"



/////////////////////////////////  Macros /////////////////////////////////////

//#ifdef _DEBUG
//#define new DEBUG_NEW
//#undef THIS_FILE
//static char THIS_FILE[] = __FILE__;
//#endif

const UINT g_nTrayTimer = 101;

///////////////////////////////// Implementation //////////////////////////////

UINT CTrayIconHooker::wm_TaskbarCreated = RegisterWindowMessage(_T("TaskbarCreated"));

LRESULT CTrayIconHooker::OnTaskbarCreated(UINT /*uMsg*/,WPARAM /*wParam*/, LPARAM /*lParam*/, BOOL& bHandled)
{
	_ASSERTE(m_pTrayIcon);
	
	//Refresh the tray icon if necessary
	if (m_pTrayIcon->IsShowing())
	{
		m_pTrayIcon->HideIcon();
		m_pTrayIcon->ShowIcon();
	}
	
	bHandled=FALSE; // allow to fall through
	return 0L;
}


CTrayIconHooker::CTrayIconHooker()
{
  //Set all of our variables to defaults
  m_pTrayIcon = NULL;
  m_phIcons = NULL;
  m_nNumIcons = 0;
  m_nTimerID = 0;
  m_nCurrentIconIndex = 0;
}

CTrayIconHooker::~CTrayIconHooker()
{
  StopUsingAnimation();
  if (m_hWnd)
	UnsubclassWindow();
}

void CTrayIconHooker::Init(CTrayNotifyIcon* pTrayIcon, CWindow* pNotifyWnd)
{
	//Validate our parameters
	_ASSERTE(pTrayIcon); //must have a valid tray notify instance
	_ASSERTE(pNotifyWnd && ::IsWindow(pNotifyWnd->m_hWnd));

	//Hook the top level frame of the notify window in preference 
	//to the notify window itself. This will ensure that we get
	//the taskbar created message
	SubclassWindow(pNotifyWnd->m_hWnd);

	m_pTrayIcon = pTrayIcon;
}


//LRESULT CTrayIconHooker::OnCreate(UINT /*uMsg*/,WPARAM /*wParam*/, LPARAM /*lParam*/, BOOL& bHandled)
//{	
//	//create the animation timer
//#if ( _ATL_VER <= 0x202 )
//	m_nTimerID = SetTimer(1, m_dwDelay, NULL); // ATL 2.2
//#elif (_ATL_VER >= 0x300 ) 
//	m_nTimerID = SetTimer(1, m_dwDelay); // ATL 3.0
//#endif
//
//	bHandled=TRUE;
//
//	return 0L;
//}

LRESULT CTrayIconHooker::OnDestroy(UINT /*uMsg*/,WPARAM /*wParam*/, LPARAM /*lParam*/, BOOL& bHandled)
{
	//kill the animation timer
	if (m_nTimerID && m_pTrayIcon->GetNotificationWnd()->m_hWnd)
		m_pTrayIcon->GetNotificationWnd()->KillTimer(m_nTimerID);
	bHandled=FALSE; // allow it to fall through
	return 0L;
}

LRESULT CTrayIconHooker::OnTimer(UINT /*uMsg*/,WPARAM wParam, LPARAM /*lParam*/, BOOL& bHandled)
{
	//increment the icon index WM_TIMER

	if ((UINT)wParam == m_nTimerID)
	{
		++m_nCurrentIconIndex;
		m_nCurrentIconIndex = m_nCurrentIconIndex % m_nNumIcons;
		
		//update the tray icon
		m_pTrayIcon->m_NotifyIconData.uFlags = NIF_ICON;
		m_pTrayIcon->m_NotifyIconData.hIcon = m_phIcons[m_nCurrentIconIndex];
		Shell_NotifyIcon(NIM_MODIFY, &m_pTrayIcon->m_NotifyIconData);
	
		bHandled=TRUE;
	}
	else
		bHandled = FALSE; // allow to fall through
	return 0L;
}

void CTrayIconHooker::StartUsingAnimation(HICON* phIcons, int nNumIcons, DWORD dwDelay)
{
  //Validate our parameters
  _ASSERTE(nNumIcons >= 2); //must be using at least 2 icons if you are using animation
  _ASSERTE(phIcons); //array of icon handles must be valid
  _ASSERTE(dwDelay); //must be non zero timer interval

  //Stop the animation if already started  
  StopUsingAnimation();

  //Hive away all the values locally
  m_phIcons = new HICON[nNumIcons];
  for (int i=0; i<nNumIcons; i++)
    m_phIcons[i] = phIcons[i];
  m_nNumIcons = nNumIcons;

  CWindow *pWnd;

  pWnd=m_pTrayIcon->GetNotificationWnd();
  if (pWnd)
  {
  //Start up the timer 
#if ( _ATL_VER <= 0x202 )
	m_nTimerID = pWnd->SetTimer(g_nTrayTimer, dwDelay, NULL); // ATL 2.2
#elif (_ATL_VER >= 0x300 ) 
	m_nTimerID = pWnd->SetTimer(g_nTrayTimer, dwDelay); // ATL 3.0
#endif
  }
}

void CTrayIconHooker::StopUsingAnimation()
{
  //Kill the timer
  if (m_nTimerID)
  {
    if (m_pTrayIcon->GetNotificationWnd()->m_hWnd)
      m_pTrayIcon->GetNotificationWnd()->KillTimer(m_nTimerID);
    m_nTimerID = 0;
  }
 
  //Free up the memory
  if (m_phIcons)
  {
    delete [] m_phIcons;
    m_phIcons = NULL;
  }

  //Reset the other animation related variables
  m_nCurrentIconIndex = 0;
  m_nNumIcons = 0;
}

HICON CTrayIconHooker::GetCurrentIcon() const 
{ 
  _ASSERTE(UsingAnimatedIcon());
  _ASSERTE(m_phIcons);
  return m_phIcons[m_nCurrentIconIndex]; 
}

BOOL CTrayIconHooker::UsingAnimatedIcon() const
{
  return (m_nNumIcons != 0);
}
			  
//IMPLEMENT_DYNAMIC(CTrayNotifyIcon, CObject)

CTrayNotifyIcon::CTrayNotifyIcon()
{
	memset(&m_NotifyIconData, 0, sizeof(m_NotifyIconData));
	m_bCreated = FALSE;
	m_bHidden = FALSE;
	m_pNotificationWnd = NULL;

 	m_uiDefaultMenuItem=0;// SAW
}

CTrayNotifyIcon::~CTrayNotifyIcon()
{
	RemoveIcon();
}

void CTrayNotifyIcon::HideIcon()
{
	_ASSERTE(m_bCreated);
	if (!m_bHidden) 
	{
		m_NotifyIconData.uFlags = NIF_ICON;
		Shell_NotifyIcon(NIM_DELETE, &m_NotifyIconData);
		m_bHidden = TRUE;
	}
}

void CTrayNotifyIcon::ShowIcon()
{
	_ASSERTE(m_bCreated);
	if (m_bHidden) 
	{
		m_NotifyIconData.uFlags = NIF_MESSAGE | NIF_ICON | NIF_TIP;
		Shell_NotifyIcon(NIM_ADD, &m_NotifyIconData);
		m_bHidden = FALSE;
	}
}

void CTrayNotifyIcon::RemoveIcon()
{
	if (m_bCreated)
	{
		m_NotifyIconData.uFlags = 0;
		Shell_NotifyIcon(NIM_DELETE, &m_NotifyIconData);
		m_bCreated = FALSE;
	}
}

void CTrayNotifyIcon::MoveToExtremeRight()
{
	HideIcon();
	ShowIcon();
}

BOOL CTrayNotifyIcon::Create(CWindow* pNotifyWnd, UINT uID, LPCTSTR pszTooltipText, HICON hIcon, UINT nNotifyMessage)
{	
	//Make sure Notification window is valid
	//ASSERT(pNotifyWnd && ::IsWindow(pNotifyWnd->GetSafeHwnd()));
	m_pNotificationWnd = pNotifyWnd;
	
	//Make sure we avoid conflict with other messages
	_ASSERTE(nNotifyMessage >= WM_USER);
	
	//Tray only supports tooltip text up to 64 characters
	_ASSERTE(_tcslen(pszTooltipText) <= 64);

	m_Menu=::LoadMenu(_Module.GetModuleInstance(),MAKEINTRESOURCE(uID));
	_ASSERTE(m_Menu);

	HMENU submenu;

	submenu=::GetSubMenu(m_Menu,0);
	_ASSERTE(submenu); // badly designed resource

	//Install the hook
	m_HookWnd.Init(this, pNotifyWnd);

	m_NotifyIconData.cbSize = sizeof(m_NotifyIconData);
	m_NotifyIconData.hWnd = pNotifyWnd->m_hWnd;
	m_NotifyIconData.uID = uID;
	m_NotifyIconData.uFlags = NIF_ICON | NIF_MESSAGE | NIF_TIP;
	m_NotifyIconData.uCallbackMessage = nNotifyMessage;
	m_NotifyIconData.hIcon = hIcon;
	_tcscpy(m_NotifyIconData.szTip, pszTooltipText);
	
	BOOL rVal = Shell_NotifyIcon(NIM_ADD, &m_NotifyIconData);
	m_bCreated = rVal;
	
	return rVal;
}


BOOL CTrayNotifyIcon::Create(CWindow* pNotifyWnd, UINT uID, LPCTSTR pszTooltipText, HICON* phIcons, int nNumIcons, DWORD dwDelay, UINT nNotifyMessage)
{
	//must be using at least 2 icons
	_ASSERTE(nNumIcons >= 2);
		
	//let the normal Create function do its stuff
	BOOL bSuccess = Create(pNotifyWnd, uID, pszTooltipText, phIcons[0], nNotifyMessage);
	
	m_HookWnd.StartUsingAnimation(phIcons, nNumIcons, dwDelay);
	
	return bSuccess;
}

BOOL CTrayNotifyIcon::SetTooltipText(LPCTSTR pszTooltipText)
{
	if (!m_bCreated)
		return FALSE;
	
	m_NotifyIconData.uFlags = NIF_TIP;
	_tcscpy(m_NotifyIconData.szTip, pszTooltipText);
	
	return Shell_NotifyIcon(NIM_MODIFY, &m_NotifyIconData);
}


BOOL CTrayNotifyIcon::SetTooltipText(UINT nID)
{
	//CString sToolTipText;
	//VERIFY(sToolTipText.LoadString(nID));
	
	TCHAR sToolTipText[65];
	
	LoadString(_Module.GetResourceInstance(),nID,sToolTipText,64);
	
	return SetTooltipText(sToolTipText);
}


BOOL CTrayNotifyIcon::SetIcon(HICON hIcon)
{
	if (!m_bCreated)
		return FALSE;

	//Since we are going to use one icon, stop any animation
	m_HookWnd.StopUsingAnimation();

	m_NotifyIconData.uFlags = NIF_ICON;
	m_NotifyIconData.hIcon = hIcon;
	
	return Shell_NotifyIcon(NIM_MODIFY, &m_NotifyIconData);
}


BOOL CTrayNotifyIcon::SetIcon(LPCTSTR lpIconName)
{
	HICON hIcon =LoadIconResource(lpIconName);
	
	return SetIcon(hIcon);
}

BOOL CTrayNotifyIcon::SetIcon(UINT nIDResource)
{
	HICON hIcon =LoadIconResource(nIDResource);
	
	return SetIcon(hIcon);
}


BOOL CTrayNotifyIcon::SetStandardIcon(LPCTSTR lpIconName)
{
	HICON hIcon = ::LoadIcon(NULL, lpIconName);
	
	return SetIcon(hIcon);
}

BOOL CTrayNotifyIcon::SetStandardIcon(UINT nIDResource)
{
	HICON hIcon = ::LoadIcon(NULL, MAKEINTRESOURCE(nIDResource));
	
	return SetIcon(hIcon);
}

BOOL CTrayNotifyIcon::SetIcon(HICON* phIcons, int nNumIcons, DWORD dwDelay)
{
	_ASSERTE(nNumIcons >= 2);
	_ASSERTE(phIcons);
	
	if (!SetIcon(phIcons[0]))
		return FALSE;

	m_HookWnd.StartUsingAnimation(phIcons,nNumIcons,dwDelay);
		
	return TRUE;
}

BOOL CTrayNotifyIcon::UsingAnimatedIcon() const
{
  return m_HookWnd.UsingAnimatedIcon();
}

CWindow* CTrayNotifyIcon::GetNotificationWnd() const
{
  return m_pNotificationWnd;
}
BOOL CTrayNotifyIcon::SetNotificationWnd(CWindow* pNotifyWnd)
{
	if (!m_bCreated)
		return FALSE;
	
	//Make sure Notification window is valid
	//ASSERT(pNotifyWnd && ::IsWindow(pNotifyWnd->GetSafeHwnd()));
	m_HookWnd.UnsubclassWindow();
	m_HookWnd.SubclassWindow(pNotifyWnd->m_hWnd);

	m_pNotificationWnd = pNotifyWnd;
	m_NotifyIconData.hWnd = pNotifyWnd->m_hWnd;
	m_NotifyIconData.uFlags = 0;
	
	return Shell_NotifyIcon(NIM_MODIFY, &m_NotifyIconData);
}

LPCTSTR CTrayNotifyIcon::GetTooltipText() const
{
	if (m_bCreated)
		return m_NotifyIconData.szTip;
	else
		return NULL;
}

HICON CTrayNotifyIcon::GetIcon() const
{
	HICON hIcon = NULL;
	if (m_bCreated)
	{
		if (UsingAnimatedIcon())
			hIcon = m_HookWnd.GetCurrentIcon();
		else
			hIcon = m_NotifyIconData.hIcon;
	}
	
	return hIcon;
}

void CTrayNotifyIcon::ModifyMenu(HMENU hMenu)
{
	if (NULL==hMenu) return;
	for (int i=0;i<GetMenuItemCount(hMenu);i++)
	{
		ModifyMenu(GetSubMenu(hMenu,i));
		UINT nID=GetMenuItemID(hMenu,i);
		m_pNotificationWnd->SendMessage(WM_COMMAND,MAKELONG(nID,-1),(LPARAM)hMenu);
	}
}

/////////////////////////////////////////////////////////////////////////////
// CTrayIcon implentation of Default Menu Item (BOLD)

void CTrayNotifyIcon::SetDefaultMenuItem(UINT uiMenuItem)
{
	m_uiDefaultMenuItem=uiMenuItem;
}

UINT CTrayNotifyIcon::GetDefaultMenuItem()
{
	return m_uiDefaultMenuItem;
}

void CTrayNotifyIcon::SetMenu(HMENU hmenu)
{
	_ASSERTE(hmenu);

	if (m_Menu)
		::DestroyMenu(m_Menu);

	m_Menu=hmenu;
}

LRESULT CTrayNotifyIcon::OnTrayNotification(WPARAM wID, LPARAM lEvent)
{
	
	//Return quickly if its not for this tray icon
	if (wID != m_NotifyIconData.uID)
		return 0L;
		
	//As a default action use a menu resource with the same id 
	//as this was created with
/*
	HMENU LoadMenu(	HINSTANCE hInstance,	// handle to application instance
					LPCTSTR lpMenuName		// menu name string or menu-resource 
											// identifier);
*/
	HMENU submenu;

	submenu=::GetSubMenu(m_Menu,0);
	_ASSERTE(submenu);
	if (!submenu) // badly designed resource
		return 0L;
	
	if (lEvent == WM_RBUTTONUP)
	{
		//Clicking with right button brings up a context menu
		ModifyMenu(submenu);
		
		// Make first menu item the default (bold font)
		::SetMenuDefaultItem(submenu, m_uiDefaultMenuItem, m_uiDefaultMenuItem? FALSE:TRUE); // SAW
		
		//Display and track the popup menu
		POINT pos;
		GetCursorPos(&pos);
		::SetForegroundWindow(m_NotifyIconData.hWnd);  
		::TrackPopupMenu(submenu, 0, pos.x, pos.y, 0, m_NotifyIconData.hWnd, NULL);

		::PostMessage(m_NotifyIconData.hWnd, WM_NULL, 0, 0);
		
	} 
	else if (lEvent == WM_LBUTTONDBLCLK) 
	{
		// double click received, the default action is to execute first menu item
		::SendMessage(m_NotifyIconData.hWnd, WM_COMMAND, m_uiDefaultMenuItem ? m_uiDefaultMenuItem : GetMenuItemID(submenu, 0), 0); // SAW
	}
	
	return 1; // handled
}

HICON CTrayNotifyIcon::LoadIconResource(LPCTSTR lpIconName)
{
  //First try to load the 16*16 icon, if this fails, they fall
  //back on the 32*32 
  HICON hIcon = (HICON) ::LoadImage(_Module.GetModuleInstance(), lpIconName, IMAGE_ICON, 16, 16, 0);
  if (hIcon == NULL)
    hIcon = LoadIcon(_Module.GetModuleInstance(),lpIconName); //AfxGetApp()->LoadIcon(lpIconName);

  //Return the icon handle
  return hIcon;
}

HICON CTrayNotifyIcon::LoadIconResource(UINT nIDResource)
{
  return LoadIconResource(MAKEINTRESOURCE(nIDResource));
}


